-- where the movesets are held
-- Hooks & Functions

-- Specki

local function speckiRoll(p)
	if p.powers[pw_carry] then return end
	if P_PlayerInPain(p) then return end
	if p.powers[pw_nocontrol] then return end
	if (p.pflags & PF_STASIS) then return end
	
    if p.specki_spintap and not (p.pflags & PF_SPINNING) and p.specki_floortouch then
		if p.speed > 6*FRACUNIT then
			p.pflags = $|PF_SPINNING|PF_THOKKED
			S_StartSound(p.mo, sfx_spin, p)
			p.mo.state = S_PLAY_ROLL
		end
	elseif p.specki_spintap and (p.pflags & PF_SPINNING) then
		if p.speed > FRACUNIT then
			p.pflags = $ & ~(PF_SPINNING|PF_THOKKED)
			p.mo.state = S_PLAY_WALK
		else
			P_InstaThrust(p.mo, p.mo.angle, 0)
			p.pflags = $ & ~(PF_SPINNING|PF_THOKKED)
			p.mo.state = S_PLAY_STND
		end
	end
	
	if (p.pflags & PF_SPINNING) and p.specki_floortouch then
		P_Thrust(p.mo, R_PointToAngle2(0, 0, p.mo.momx, p.mo.momy), -p.mo.friction)
	elseif (p.pflags & PF_SPINNING) and (p.pflags & PF_JUMPED) then
		p.pflags = $ & ~PF_SPINNING
	end
	
	if p.pflags & PF_SPINNING then
		local thok = P_SpawnGhostMobj(p.mo)
		if p.specki_rainboweffect then
			thok.color = leveltime % 68
		else
			thok.color = p.mo.color
		end
		thok.sprite = SPR_THOK
		thok.frame = A
		thok.scale = FRACUNIT+FRACUNIT/6
		thok.z = $ - p.mo.height*P_MobjFlip(p.mo)/14
		thok.destscale = FRACUNIT/32
		thok.blendmode = AST_ADD
		thok.fuse = 6
		
		P_ButteredSlope(p.mo)
		if leveltime % 8 == 0 then
			P_ButteredSlope(p.mo)
		end
	end
end

local function speckiAccel(p)
	if p.specki_runtimer == nil then
		p.specki_runtimer = 0
	end

	if p.speed >= p.runspeed and p.specki_floortouch and not (p.pflags & PF_SPINNING) then		
		if p.mo.eflags & MFE_SPRUNG then
			p.specki_runtimer = $ + 105
		end
		
		if p.powers[pw_sneakers] then
			p.specki_runtimer = $ + 1
		end

		if leveltime % 3 == 0 then
			p.specki_runtimer = $ + 2
		end
		
		if p.specki_runtimer > 16*TICRATE then
			p.specki_runtimer = 16*TICRATE
		end
		
		if (p.cmd.forwardmove == 0 and p.cmd.sidemove == 0) then
			p.specki_runtimer = 0
		end
		
		if p.mo.friction < FRACUNIT then 
			P_Thrust(p.mo, R_PointToAngle2(0, 0, p.rmomx, p.rmomy), p.specki_runtimer / 20 * p.mo.friction/8)
		else
			P_Thrust(p.mo, R_PointToAngle2(0, 0, p.rmomx, p.rmomy), p.specki_runtimer / 20 * FRACUNIT)
		end
	elseif (p.speed < p.runspeed) and (p.specki_runtimer > 0) and not (p.pflags & PF_SPINNING) then
		p.specki_runtimer = $ - 6
	end
	
	if p.specki_runtimer > 0 then
		local ghost = P_SpawnGhostMobj(p.mo)
		if p.specki_rainboweffect then
			ghost.color = leveltime % 68
		else
			ghost.color = p.mo.color
		end
		ghost.colorized = true
		ghost.blendmode = AST_ADD
		ghost.fuse = 2
	end
	
	if (p.pflags & PF_SPINNING) or not p.specki_floortouch then
		if p.specki_runtimer > 0 then
			p.specki_runtimer = $ - p.specki_runtimer / TICRATE
		end
	end
	
	if P_PlayerInPain(p) then
		p.specki_runtimer = 0
	end
end

local function speckiBounce(p)
	if p.powers[pw_shield] then return end

	if (p.pflags & PF_JUMPED) and p.specki_spintap and not p.specki_bouncing then
		p.specki_bouncing = true
		p.mo.state = S_PLAY_ROLL
		P_SetObjectMomZ(p.mo, -32*p.jumpfactor, false)
		p.powers[pw_strong] = STR_ANIM|STR_FLOOR|STR_SPRING
	elseif p.specki_floortouch and (p.cmd.buttons & BT_USE) and p.specki_bouncing then
		p.specki_bouncing = false
		p.specki_trickstimer = 0
		p.mo.state = S_PLAY_ROLL
		p.pflags = $|PF_JUMPED & ~PF_SPINNING
		if abs(p.specki_lastmomz/16*6) < 14*p.jumpfactor then
			P_SetObjectMomZ(p.mo, 14*p.jumpfactor, false)
		else
			P_SetObjectMomZ(p.mo, -(p.specki_lastmomz/16*6), false)
		end
	elseif p.specki_floortouch and p.specki_bouncing then
		p.specki_bouncing = false
		p.specki_trickstimer = 0
		p.mo.state = S_PLAY_RUN
		p.pflags = $ & ~(PF_JUMPED|PF_SPINNING)
	end
	
	if p.specki_bouncing then
		P_SetObjectMomZ(p.mo, -gravity, true)
		
		local thok = P_SpawnGhostMobj(p.mo)
		if p.specki_rainboweffect then
			thok.color = leveltime % 68
		else
			thok.color = p.mo.color
		end
		thok.sprite = SPR_THOK
		thok.frame = A
		thok.scale = FRACUNIT+FRACUNIT/6
		thok.z = $ - p.mo.height*P_MobjFlip(p.mo)/14
		thok.destscale = FRACUNIT/32
		thok.blendmode = AST_ADD
		thok.fuse = 6
		
		P_ButteredSlope(p.mo)
		if leveltime % 8 == 0 then
			P_ButteredSlope(p.mo)
		end
	end
	
	if P_PlayerInPain(p) or p.specki_sprung or p.specki_ingoop or (p.deathpitlaunchcooldown ~= nil and p.deathpitlaunchcooldown <= 0) then
		p.specki_bouncing = false
	end
end

addHook("MobjDeath", function(emo, inf, src)
	if not (emo and emo.valid) then return end
	if not ((emo.flags & MF_ENEMY) or (emo.flags & MF_BOSS) or (emo.flags & MF_MONITOR)) then return end
	if not (src and src.valid) then return end
	if not src.health then return end
	if src.type ~= MT_PLAYER then return end
	if src.skin ~= "specki" then return end
	if src.player.specki_bouncing then
		src.momz = $/3*2
		src.player.specki_bouncing = false
		src.player.specki_trickstimer = 3*TICRATE
	end
end)

local function speckiSleep(p)
	if p.specki_sleeptimer == nil then
		p.specki_sleeptimer = 0
	end
		
	if p.mo.state == S_SPECKI_SLEEP then
		p.specki_sleeptimer = $+1
	end
		
	if p.mo.state == S_PLAY_WAIT then
		p.mo.frame = (leveltime/6)%3
		p.specki_sleeptimer = $+1
			
		if p.specki_sleeptimer >= 5*TICRATE then
			p.mo.frame = 1
				
			if p.specki_sleeptimer >= 5*TICRATE+8 then
				p.mo.state = S_SPECKI_SLEEP
			end
		end
	elseif p.mo.state ~= S_SPECKI_SLEEP and p.mo.state ~= S_PLAY_WAIT then
		p.specki_sleeptimer = 0
	end
end

local function speckiTaunt(p)
	if p.specki_tosstap and p.specki_floortouch and p.mo.state ~= S_SPECKI_DANCE and p.mo.state ~= S_SPECKI_PRIDE and not p.speed then
		if (p.cmd.buttons & BT_CUSTOM3) then
			p.mo.state = S_SPECKI_PRIDE
			S_StartSound(p.mo, sfx_cdpcm6)
		else
			p.mo.state = S_SPECKI_DANCE
		end
	end
	
	if p.mo.state == S_SPECKI_DANCE or p.mo.state == S_SPECKI_SLEEP or p.mo.state == S_SPECKI_PRIDE then
		if p.speed ~= 0 then
			p.mo.state = S_PLAY_STND
		elseif p.mo.state == S_SPECKI_DANCE then
			if S_MusicName(p) ~= "FRIDAY" then
				p.ospeckimusic = S_MusicName(p)
				S_ChangeMusic("FRIDAY", true, p)
			end
		end
	end
	
	if p.specki_laststate == S_SPECKI_DANCE and p.mo.state ~= S_SPECKI_DANCE then
		if string.lower(p.ospeckimusic) == "_super" and not p.powers[pw_super] then
			p.ospeckimusic = mapheaderinfo[gamemap].musname
		end
		
		S_ChangeMusic(p.ospeckimusic, true, p)
	end
end

local function speckiTricks(p)
	if p.specki_trickstimer == nil then
		p.specki_trickstimer = 0
	end
	
	if P_PlayerInPain(p) or (p.deathpitlaunchcooldown ~= nil and not p.deathpitlaunchcooldown) or p.specki_floortouch > 2 then 
		p.specki_trickstimer = 0
		return 
	end

	if (p.mo.eflags & MFE_SPRUNG)
	or (p.powers[pw_carry] == CR_MACESPIN) 
	or (p.powers[pw_carry] == CR_MINECART) then
		p.specki_trickstimer = 3*TICRATE
	end
	
	if p.specki_tricks == nil then
		p.specki_tricks = 0
	end
	
	if p.specki_tricks > 0 and not p.linktimer then
		p.specki_tricks = 0
	end
	
	if p.specki_tricks == 30 then
		speckichars.giveachievement(2)
		P_AddPlayerScore(p, 10000)
		p.specki_tricks = $ + 1
	end
	
	if p.powers[pw_carry] then return end
	
	if not p.specki_floortouch 
	and p.specki_trickstimer 
	and p.specki_jumptap 
	and p.mo.state ~= S_SPECKI_TRICK then
		p.mo.state = S_SPECKI_TRICK
		p.mo.sprite2 = speckichars.specki_trickframes[P_RandomRange(1, #speckichars.specki_trickframes)][1]
		p.mo.frame = speckichars.specki_trickframes[P_RandomRange(1, #speckichars.specki_trickframes)][2]
		
		P_DoNightsScore(p)
		p.specki_tricks = $+1
		
		local ghost = P_SpawnGhostMobj(p.mo)
		ghost.blendmode = AST_ADD
		ghost.renderflags = RF_FULLBRIGHT
		if p.specki_rainboweffect then
			ghost.color = leveltime % 68
		else
			ghost.color = p.mo.color
		end
		ghost.colorized = true
		
		local thok = P_SpawnGhostMobj(p.mo)
		thok.blendmode = AST_ADD
		thok.renderflags = RF_FULLBRIGHT
		if p.specki_rainboweffect then
			thok.color = leveltime % 68
		else
			thok.color = p.mo.color
		end
		thok.sprite = SPR_THOK
		thok.frame = A
		thok.scale = 2*p.mo.scale
		
		p.specki_trickstimer = $-8
	end
	
	if p.specki_trickstimer > 0 then
		local ghost = P_SpawnGhostMobj(p.mo)
		ghost.blendmode = AST_ADD
		ghost.renderflags = RF_FULLBRIGHT
		if p.specki_rainboweffect then
			ghost.color = leveltime % 68
		else
			ghost.color = p.mo.color
		end
		ghost.colorized = true
		ghost.fuse = 3
		
		p.specki_trickstimer = $-1
	else
		p.specki_trickstimer = 0
	end
end

-- Maril

local function marilPiano(p)
	if p.specki_tosstap and p.specki_floortouch and not p.speed then
		if not p.maril_piano then
			p.maril_piano = true
			p.mo.sprite2 = SPR2_MLEE
			p.mo.frame = A
			p.powers[pw_nocontrol] = 15*TICRATE
		else
			p.maril_piano = false
			p.mo.sprite2 = SPR2_STND
			p.mo.frame = A
			p.powers[pw_nocontrol] = 0
			return
		end
	end
	
	if p.maril_piano then
		p.mo.sprite2 = SPR2_MLEE
		p.powers[pw_nocontrol] = 15*TICRATE
	end
end

local function marilOuttaHere(p)
	if p.mo.state == S_PLAY_WAIT and not (splitscreen or multiplayer or netgame) and p.maril_waittimer ~= nil then
		p.maril_waittimer = $ + 1
		
		if p.maril_waittimer >= 178*TICRATE then
			p.mo.state = S_PLAY_MARIL_OUTTAHERE
			p.playerstate = PST_DEAD
			p.lives = 0
			P_SetObjectMomZ(p.mo, 10*p.jumpfactor, false)
		end
	else
		p.maril_waittimer = 0
	end
end

local function marilSpin(p)
	if p.specki_spintap and not (p.pflags & PF_THOKKED) and not (p.maril_spin or p.specki_sprung or p.maril_piano) then
		p.maril_spin = true 
		p.pflags = $|PF_THOKKED
		p.mo.state = S_PLAY_MARIL_SPIN
		
		S_StartSound(p.mo, sfx_spspin)
		
		if not p.specki_floortouch then
			P_SetObjectMomZ(p.mo, 8*p.jumpfactor, false)
		elseif p.speed > FRACUNIT then
			P_Thrust(p.mo, R_PointToAngle2(0, 0, p.mo.momx, p.mo.momy), p.thrustfactor<<FRACBITS*2)
		end
	end
		
	if p.mo.state == S_PLAY_MARIL_SPIN then
		local ghost = P_SpawnGhostMobj(p.mo)
		ghost.fuse = 3
		p.powers[pw_strong] = STR_ANIM|STR_ATTACK|STR_WALL|STR_FLOOR|STR_SPRING
	end
		
	if p.mo.state ~= S_PLAY_MARIL_SPIN and p.maril_spin then
		p.pflags = $ & ~PF_SPINNING
		
		if p.specki_floortouch then
			p.pflags = $ & ~PF_THOKKED
			
			if (p.cmd.buttons & BT_USE) then
				p.mo.state = S_PLAY_ROLL
				p.pflags = $|PF_SPINNING
				S_StartSound(p.mo, sfx_spin)
			end
		end
		
		p.maril_spin = false
	end
end

addHook("PlayerCanDamage", function(p)
	if not (p.mo and p.mo.valid) then return end
	if not p.mo.health then return end
	if p.mo.skin ~= "maril" then return end
	if p.mo.state == S_PLAY_MARIL_SPIN then return true end
	if p.mo.state == S_PLAY_MARIL_SWIM_SPIN then return true end
end)

addHook("PlayerHeight", function(p)
	if not (p.mo and p.mo.valid) then return end
	if not p.mo.health then return end
	if p.mo.skin ~= "maril" then return end
	if p.mo.state == S_PLAY_MARIL_SPIN or (p.pflags & PF_SPINNING) then
		return P_GetPlayerSpinHeight(p)
	else
		return P_GetPlayerHeight(p)
	end
end)

addHook("PlayerCanEnterSpinGaps", function(p)
	if not (p.mo and p.mo.valid) then return end
	if not p.mo.health then return end
	if p.mo.skin ~= "maril" then return end
	if p.mo.state == S_PLAY_MARIL_SPIN then return true end
end)

addHook("ShouldDamage", function(maril, inf, spike)
	if not (maril and maril.valid) then return end
	if not maril.health then return end
	if maril.skin ~= "maril" then return end
	if not (spike and spike.valid) then return end
	if spike.type ~= MT_SPIKE then return end
	if (spike.flags2 & MF2_OBJECTFLIP) and not (maril.flags2 & MF2_OBJECTFLIP) then return end
	if not (spike.flags2 & MF2_OBJECTFLIP) and (maril.flags2 & MF2_OBJECTFLIP) then return end
	if maril.player.maril_spin == true then return end
	if maril.state ~= S_PLAY_ROLL then return end
	if not (maril.player.pflags & PF_THOKKED) then return end
	if maril.player.cmd.buttons & BT_JUMP then
		P_SetObjectMomZ(maril, 9*maril.player.jumpfactor, false)
	else
		P_SetObjectMomZ(maril, 6*maril.player.jumpfactor, false)
	end
	P_SpawnMobj(maril.x, maril.y, (maril.z+maril.height/2*P_MobjFlip(maril)) - (P_MobjFlip(maril)*maril.height/2), MT_SUPERSPARK)
	return false
end, MT_PLAYER)

addHook("AbilitySpecial", function(p)
	if not (p.mo and p.mo.valid) then return end
	if not p.mo.health then return end
	if p.mo.skin ~= "maril" then return end
	if mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS then return end
	if mapheaderinfo[gamemap].typeoflevel & TOL_XMAS then return end
	if p.pflags & PF_THOKKED then return end
	
	if not p.maril_fly then
		p.maril_fly = true
	end
end)

addHook("ShouldDamage", function(enemy, inf, pmo)
	if not (enemy and enemy.valid) then return end
	if not enemy.health then return end
	if not (enemy.flags & MF_ENEMY) then return end
	if not (pmo and pmo.valid) then return end
	if pmo.type ~= MT_PLAYER then return end
	local p = pmo.player
	if pmo.skin ~= "maril" then return end
	if pmo.state ~= S_PLAY_MARIL_SPIN then return end
	if enemy.swiped then return false end
	if enemy.kitsune == pmo then return false end
	
	if not enemy.swiped then
		enemy.swiped = pmo
	end
	
	S_StartSoundAtVolume(enemy, sfx_cdfm28, 255)
	P_Thrust(enemy, R_PointToAngle2(0, 0, pmo.momx, pmo.momy), pmo.player.speed/2)
	enemy.flags = $ & ~MF_NOGRAVITY
	P_SetObjectMomZ(enemy, 6*FRACUNIT + pmo.player.speed/5, false)
	enemy.z = $+(16*FRACUNIT)*P_MobjFlip(enemy)
	
	return false
end)

addHook("ShouldDamage", function(player, inf, enemy)
	if not (enemy and enemy.valid) then return end
	if not enemy.health then return end
	if not (enemy.flags & MF_ENEMY) then return end
	if not (player and player.valid) then return end
	if not player.health then return end	
	if enemy.swiped then return false end
end, MT_PLAYER)

addHook("MobjThinker", function(enemy)
	if not (enemy and enemy.valid) then return end
	if not enemy.health then return end
	if not enemy.swiped then return end
	if P_IsObjectOnGround(enemy) then
		enemy.rollangle = 0
		P_DamageMobj(enemy, enemy.swiped, enemy.swiped)
		enemy.swiped = nil
		return
	end
	
	if enemy.state ~= enemy.info.spawnstate then
		P_SetMobjStateNF(enemy, enemy.info.spawnstate)
	end
	
	if enemy.turnangle == nil then
		enemy.turnangle = FixedAngle(12*FRACUNIT)
	else
		enemy.turnangle = $+FixedAngle(12*FRACUNIT)
	end
	
	local ghost = P_SpawnGhostMobj(enemy)
	ghost.color = enemy.swiped.color
	ghost.colorized = true
	ghost.scale = enemy.scale/3*2
	
	enemy.rollangle = enemy.turnangle
	enemy.angle = enemy.turnangle
end)

addHook("MobjMoveBlocked", function(enemy, thing, line)
	if not (enemy and enemy.valid) then return end
	if not enemy.health then return end
	if not enemy.swiped then return end

	if line and line.valid
		if FixedHypot(enemy.momx, enemy.momy) > 4*FRACUNIT
			if abs(line.dx) > 0
				local myang = R_PointToAngle2(0, 0, enemy.momx, enemy.momy)
				local vertang = R_PointToAngle2(0, 0, 0, enemy.momz)
				local lineang = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
				P_InstaThrust(enemy, myang + 2*(lineang - myang), FixedHypot(enemy.momx, enemy.momy))
				if not thing then S_StartSoundAtVolume(mo, sfx_bnce2, 127) end
			else
				enemy.momx = $*-1
			end
			
			return true
		else
			P_DamageMobj(enemy, enemy.swiped, enemy.swiped, 1)
		end
	end
end)

addHook("MobjMoveCollide", function(enemy, thing)
	if not (enemy and enemy.valid) then return end
	if not enemy.health then return end
	if not enemy.swiped then return end
	if not (thing and thing.valid) then return end
	if not thing.health then return end 
	if not (thing.flags & MF_ENEMY) then return end
	if enemy.z > thing.z + thing.height then return end
	if enemy.z + enemy.height < thing.z then return end
	if thing.swiped then return end

	local myang = R_PointToAngle2(0, 0, enemy.momx, enemy.momy)
	P_InstaThrust(enemy, myang + 2*(thing.angle - myang), FixedHypot(enemy.momx, enemy.momy))
	
	if enemy.momz*P_MobjFlip(enemy) <= 0 then
		P_SetObjectMomZ(enemy, 6*FRACUNIT+FixedHypot(enemy.momx, enemy.momy)/5, false)
	end
	
	thing.swiped = enemy.swiped
	S_StartSoundAtVolume(thing, sfx_cdfm28, 255)
	
	P_InstaThrust(thing, myang, FixedHypot(enemy.momx, enemy.momy))
	if thing.momz*P_MobjFlip(thing) <= 0 then
		P_SetObjectMomZ(thing, 6*FRACUNIT+FixedHypot(enemy.momx, enemy.momy)/5, false)
	end
	thing.flags = $ & ~MF_NOGRAVITY
	thing.z = $+(16*FRACUNIT)*P_MobjFlip(thing)
	
	S_StartSoundAtVolume(mo, sfx_bnce2, 127)
			
	return true
end)

local function marilFly(p)
	if (p.specki_floortouch or P_PlayerInPain(p) or (p.mo.eflags & MFE_UNDERWATER) or p.specki_sprung or p.specki_sliding) then
		p.maril_fly = false
			
		if p.specki_floortouch then
			p.mo.state = S_PLAY_STND
		elseif (p.mo.eflags & MFE_UNDERWATER) then
			p.mo.state = S_PLAY_FALL
		end
			
		return
	end
	
	p.pflags = $ & ~(PF_JUMPED|PF_SPINNING)
	p.mo.state = S_PLAY_FLY
	p.mo.frame = (leveltime/4) % 2
	
	if p.cmd.buttons & BT_JUMP then
		P_SetObjectMomZ(p.mo, FRACUNIT, true)
	elseif p.cmd.buttons & BT_USE then
		P_SetObjectMomZ(p.mo, -FRACUNIT, true)
	else
		P_SetObjectMomZ(p.mo, -p.mo.momz/TICRATE, true)
	end
		
	if (p.cmd.buttons & BT_JUMP) and (p.cmd.buttons & BT_USE) then
		p.maril_fly = false
		p.pflags = $|PF_SPINNING|PF_THOKKED
		p.mo.state = S_PLAY_ROLL
		p.powers[pw_strong] = STR_ANIM|STR_ATTACK|STR_WALL|STR_FLOOR|STR_SPRING
	end
end

local function marilSwim(p)
	if (p.specki_sprung or p.powers[pw_carry] or p.specki_sliding) then return end

	p.maril_fly = false
	p.maril_spin = false
		
	p.pflags = $ & ~PF_JUMPED
		
	if p.mo.state ~= S_PLAY_MARIL_SWIM_IDLE and p.mo.state ~= S_PLAY_MARIL_SWIM and p.mo.state ~= S_PLAY_MARIL_SWIM_SPIN and not p.specki_floortouch then
		p.mo.state = S_PLAY_MARIL_SWIM_IDLE
	end
		
	if p.maril_swimspeed > 12*FRACUNIT then
		p.maril_swimspeed = 12*FRACUNIT
	end
		
	if p.maril_vertswimangle < -145 then
		p.maril_vertswimangle = -145
	elseif p.maril_vertswimangle > 145 then
		p.maril_vertswimangle = 145
	end
		
	if p.maril_swimspincooldown then
		p.maril_swimspincooldown = $ - 1
	end
		
	if p.powers[pw_flashing] < 70 then
		if p.mo.state == S_PLAY_MARIL_SWIM then
			p.maril_swimangle = $-p.cmd.sidemove/24*ANG1
		else
			p.maril_swimangle = $-p.cmd.sidemove/35*ANG1
		end
		
		if p.mo.state == S_PLAY_MARIL_SWIM then
			p.maril_vertswimangle = $-p.cmd.forwardmove/24
		else
			p.maril_vertswimangle = $-p.cmd.forwardmove/35
		end
				
		if not abs(p.cmd.forwardmove) and not p.maril_swimspeed then
			p.maril_vertswimangle = $ - $/TICRATE
		end
			
		if p.specki_spintap and p.mo.state ~= S_PLAY_MARIL_SWIM_SPIN and not p.maril_swimspincooldown and not (p.pflags & PF_SLIDING) then
			p.mo.state = S_PLAY_MARIL_SWIM_SPIN
			p.powers[pw_strong] = STR_ANIM|STR_ATTACK|STR_FLOOR|STR_WALL
			p.maril_swimspeed = 24*FRACUNIT
			p.maril_swimspincooldown = 2*TICRATE
		end
		
		P_InstaThrust(p.mo, p.maril_swimangle, p.maril_swimspeed)
		P_Thrust(p.mo, R_PointToAngle2(0, 0, p.maril_swimoldmomx, p.maril_swimoldmomy), FixedHypot(p.maril_swimoldmomx, p.maril_swimoldmomy))
		p.maril_swimoldmomx = $-$/12
		p.maril_swimoldmomy = $-$/12
		if p.specki_ingoop and abs(p.mo.momz) <= 16*FRACUNIT then
			P_SetObjectMomZ(p.mo, p.maril_swimspeed*p.maril_vertswimangle/700, true)
		else
			P_SetObjectMomZ(p.mo, p.maril_swimspeed*p.maril_vertswimangle/70+p.maril_swimoldmomz, false)
			p.maril_swimoldmomz = $-$/17
		end
			
		if p.maril_swimspeed > 16*FRACUNIT then
			p.maril_swimspeed = $ - FRACUNIT/2
		else
			p.maril_swimspeed = $ - FRACUNIT/8
		end
			
		if not ((p.pflags & PF_ANALOGMODE) and (p.pflags & PF_DIRECTIONCHAR)) then
			if p.mo.angle ~= p.maril_swimangle then
				p.mo.angle = $ - ($ - p.maril_swimangle)/17
			end
		end
			
		if p.aiming ~= p.maril_vertswimangle*ANG1 then
			p.aiming = $ - ($ - p.maril_vertswimangle*ANG1)/17
		end
			
		p.drawangle = p.maril_swimangle
	else
		p.maril_swimspeed = 0
		P_InstaThrust(p.mo, p.maril_swimangle, -6*FRACUNIT) 
		p.mo.momz = 0
	end
		
	if p.maril_swimspeed < 0 then
		p.maril_swimspeed = 0
	end
end

addHook("JumpSpecial", function(p)
	if not (p.mo and p.mo.valid) then return end
	if not p.mo.health then return end
	if p.mo.skin ~= "maril" then return end
	if not (p.mo.eflags & MFE_UNDERWATER) then return end
	if not p.maril_canswim then return end
	if mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS then return end
	if mapheaderinfo[gamemap].typeoflevel & TOL_XMAS then return end
	if (p.powers[pw_carry] or p.specki_sliding or p.specki_sprung) then return end
	
	if p.specki_jumptap and (p.specki_floortouch or p.mo.state == S_PLAY_MARIL_SWIM_IDLE or (p.mo.state == S_PLAY_MARIL_SWIM and p.mo.frame > 1)) then
		p.mo.state = S_PLAY_MARIL_SWIM
		p.powers[pw_strong] = STR_ANIM|STR_HEAVY
		p.maril_swimspeed = $+4*FRACUNIT
		P_SpawnSkidDust(p, p.mo.radius/2, false)	
	elseif not (p.specki_jumptap or p.specki_jumpready) and (p.specki_floortouch or p.mo.state == S_PLAY_MARIL_SWIM_IDLE or (p.mo.state == S_PLAY_MARIL_SWIM and p.mo.frame > 1)) then
		if leveltime % 6 == 0 then
			p.mo.state = S_PLAY_MARIL_SWIM
			p.powers[pw_strong] = STR_ANIM|STR_HEAVY
			p.maril_swimspeed = $+FRACUNIT
			P_SpawnSkidDust(p, p.mo.radius/4, false)
		end
	end
	
	return true
end)

-- Shared

addHook("MobjMoveBlocked", function(mo, thing, line)
	if not (mo and mo.valid) then return end
	if not mo.health then return end
	if not speckichars.isSpeckiChar(mo.skin) then return end
	if mo.player.specki_nobounce then return end
	if not (mo.player.pflags & PF_SPINNING) and not ((mo.player.pflags & PF_JUMPED) and mo.state == S_PLAY_ROLL) then return end

	if line and line.valid
		if FixedHypot(mo.momx, mo.momy) > 5*FRACUNIT
			if abs(line.dx) > 0
				local myang = R_PointToAngle2(0, 0, mo.momx, mo.momy)
				local vertang = R_PointToAngle2(0, 0, 0, mo.momz)
				local lineang = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
				P_InstaThrust(mo, myang + 2*(lineang - myang), FixedHypot(mo.momx, mo.momy)-mo.friction)
			else
				mo.momx = $*-1
			end
		end
	end
end, MT_PLAYER)

addHook("ThinkFrame", do
	for player in players.iterate do
		if (player.pflags & PF_SLIDING) then
			player.specki_sliding = true
		else
			player.specki_sliding = false
		end
	end
end)

local function speckiSquishy(p)
	if (p.mo.eflags & MFE_JUSTHITFLOOR) and (not (p.mo.eflags & MFE_UNDERWATER)) then
		if not p.specki_floortouch then
			p.mo.spritexscale = FRACUNIT/3*4
			p.mo.spriteyscale = FRACUNIT/3*2
		end
	elseif not P_IsObjectOnGround(p.mo) then
		if p.specki_floortouch and (p.pflags & PF_JUMPED) then
			p.mo.spritexscale = FRACUNIT/3*2
			p.mo.spriteyscale = FRACUNIT/3*4
		end
	end
	
	p.mo.spritexscale = ease.outquint(FRACUNIT/17, $, FRACUNIT)
	p.mo.spriteyscale = ease.outquint(FRACUNIT/17, $, FRACUNIT)
end

local function toboolean(v)
	if v == nil then return false end

	if type(v) == "number" then
		if v > 0 then
			return true
		else
			return false
		end
	elseif type(v) == "string" then
		if v == "true" then
			return true
		elseif v == "false" then
			return false
		else
			return true
		end
	end
end

COM_AddCommand("speckichars_load", function(p)
	local savef = io.openlocal("client/speckichars/save.txt", "r")
	local cvarf = io.openlocal("client/speckichars/cvars.txt", "r")
	local linenum = 1
	local table = {}
	local cvartable = {}
	local laterversion = false
	local tooold = false
	
	if savef == nil then
		return
	end
	
	for line in savef:lines() do
		if linenum == 1 then
			linenum = $+1
			continue
		elseif linenum == 2 then
			if tonumber(line) > speckichars.VERSIONNUM then
				laterversion = true
				break
			elseif tonumber(line) < 3 then
				tooold = true
				break
			else
				linenum = $+1
				continue
			end
		end
		
		table[linenum-2] = line
		linenum = $+1
	end
	
	savef:close()
	
	if laterversion then
		CONS_Printf(p, "this speckichars save is for a later version!! please update your speckichars!!\n")
		return
	end
	
	if tooold then
		CONS_Printf(p, "this speckichars save is for an older incompatible version!! i'm sorry!!\n")
		return
	end
	
	for i, str in ipairs(table) do
		for k, v in str:gmatch"%s*([^=]*) = ([^=]*)%f[%s%z]" do 
			speckichars.vars[k] = toboolean(v) 
			--print(k .. " = " .. v)
		end
	end
	
	if cvarf == nil then
		return
	end
	
	linenum = 1
	laterversion = false
	tooold = false
	
	for line in cvarf:lines() do
		if linenum == 1 then
			linenum = $+1
			continue
		elseif linenum == 2 then
			if tonumber(line) > speckichars.VERSIONNUM then
				laterversion = true
				break
			else
				linenum = $+1
				continue
			end
		end
		
		cvartable[linenum-2] = tonumber(line)
		linenum = $+1
	end
	
	cvarf:close()
	
	if laterversion then
		CONS_Printf(p, "this speckichars cvars file is for a later version!! please update your speckichars!!\n")
		return
	end
	
	p.openedspeckichars_menu = true
	
	local mus = CV_FindVar("speckichars_custommusic")
	local ptsrmus = CV_FindVar("speckichars_ptsrcustommusic")
	
	CV_Set(mus, cvartable[1])
	CV_Set(ptsrmus, cvartable[2])
end)

local function resetSpeckiCharsVars(p)
	if not (p.mo and p.mo.valid) then return end
	if not p.mo.health then return end

	p.specki_jumptap = false
	p.specki_jumpready = true
	p.specki_spintap = false
	p.specki_spinready = true
	p.specki_tosstap = false
	p.specki_tossready = true
	
	p.specki_floortouch = 0
	p.specki_sprung = false
	p.specki_sliding = false
	p.specki_ingoop = false
	
	p.specki_laststate = p.mo.state
	p.specki_lastmomz = p.mo.momz
	p.specki_nohit = true
	
	-- specki specific
	
	p.specki_runtimer = 0
	p.specki_sleeptimer = 0
	p.specki_bouncing = false
	p.specki_trickstimer = 0
	p.specki_rainboweffect = true
	
	-- maril specific
	
	p.maril_piano = false
	p.maril_fly = false
	p.maril_spin = false
	p.maril_canswim = true
	p.maril_swimspeed = 0
	p.maril_swimoldmomx = 0
	p.maril_swimoldmomy = 0
	p.maril_swimoldmomz = 0
	p.maril_swimangle = p.drawangle
	p.maril_vertswimangle = 0
	p.maril_waittimer = 0
	
	COM_BufInsertText(p, "speckichars_load")
	speckichars.save()
	speckichars.savecvars()
end

COM_AddCommand("speckichars_updatevars", function(p)
	p.maril_canswim = speckichars.vars.maril_canswim
	p.specki_rainboweffect = speckichars.vars.specki_rainboweffect
	
	speckichars.save(p)
end)

local function updateSpeckiCharsVars(p)
	if p.maril_canswim ~= speckichars.vars.maril_canswim
	or p.specki_rainboweffect ~= speckichars.vars.specki_rainboweffect then
		COM_BufInsertText(p, "speckichars_updatevars")
	end
end

addHook("MapChange", function(map)
	for p in players.iterate do
		if p.specki_floortouch ~= nil then
			speckichars.save()
			speckichars.savecvars()
		end
		resetSpeckiCharsVars(p)
	end
end)

addHook("PlayerThink", function(p)
	if not (p.mo and p.mo.valid) then return end
	
	if p.mo.state == S_PLAY_MARIL_OUTTAHERE then 
		P_SetObjectMomZ(p.mo, -gravity, true)
		p.mo.flags = $|MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT
		return 
	end
	
	if not p.mo.health then return end
	if not speckichars.isSpeckiChar(p.mo.skin) then return end
	
	if p.specki_floortouch == nil then
		resetSpeckiCharsVars(p)
	end
	
	updateSpeckiCharsVars(p)
	
	if not (p.cmd.buttons & BT_JUMP) then
		p.specki_jumptap = false
		p.specki_jumpready = true
	elseif p.specki_jumpready then
		p.specki_jumptap = true
		p.specki_jumpready = false
	else
		p.specki_jumptap = false
	end
	
	if not (p.cmd.buttons & BT_USE) then
		p.specki_spintap = false
		p.specki_spinready = true
	elseif p.specki_spinready then
		p.specki_spintap = true
		p.specki_spinready = false
	else
		p.specki_spintap = false
	end
	
	if not (p.cmd.buttons & BT_TOSSFLAG) then
		p.specki_tosstap = false
		p.specki_tossready = true
	elseif p.specki_tossready then
		p.specki_tosstap = true
		p.specki_tossready = false
	else
		p.specki_tosstap = false
	end

	p.powers[pw_underwater] = 9999 -- they can't drown
	p.powers[pw_spacetime] = 9999

	speckiSquishy(p)
	
	if P_IsObjectOnGround(p.mo) or (p.mo.eflags & MFE_JUSTHITFLOOR) then -- floor checks (MIDIMan pointed out a bug here where I used pflags instead of eflags, ty)
		p.specki_floortouch = $ + 1
	else
		p.specki_floortouch = 0
	end
	
	if p.mo.eflags & MFE_SPRUNG then
		p.specki_sprung = true
	elseif p.specki_floortouch or P_PlayerInPain(p) then
		p.specki_sprung = false
	end
	
	if p.mo.eflags & MFE_GOOWATER then
		p.specki_ingoop = true
	else
		p.specki_ingoop = false
	end
	
	if P_PlayerInPain(p) and (p.deathpitlaunchcooldown ~= nil and not p.deathpitlaunchcooldown) and not (p.powers[pw_shield] & SH_FORCE) then
		p.specki_nohit = false
	end

	if p.mo.skin == "specki" then	
		speckiAccel(p)
		speckiRoll(p)
		speckiBounce(p)
		speckiSleep(p)
		speckiTaunt(p)
		speckiTricks(p)
	else
		marilPiano(p)
		
		if P_PlayerInPain(p) then
			p.maril_fly = false
			p.maril_spin = false
		end
				
		if p.specki_sprung and p.maril_spin then
			p.mo.state = S_PLAY_ROLL
			p.maril_spin = false
			p.pflags = $|PF_SPINNING & ~PF_THOKKED
			S_StartSound(p.mo, sfx_spin)
		end
				
		if (p.mo.eflags & MFE_UNDERWATER) then
			if p.maril_canswim then
				marilSwim(p)
			else
				marilSpin(p)
			end
		else
			if p.mo.state == S_PLAY_MARIL_SWIM_IDLE or p.mo.state == S_PLAY_MARIL_SWIM or p.mo.state == S_PLAY_MARIL_SWIM_SPIN then
				P_DoJump(p)
				p.maril_swimspeed = 0
				p.maril_swimangle = p.drawangle
				p.maril_vertswimangle = 0
				p.maril_swimspincooldown = 0
				p.maril_postswimadjust = 24
			else
				p.maril_swimspeed = 0
				p.maril_swimangle = p.drawangle
				p.maril_vertswimangle = 0
				p.maril_swimspincooldown = 0
			end
			
			p.maril_swimoldmomx = p.mo.momx
			p.maril_swimoldmomy = p.mo.momy
			p.maril_swimoldmomz = p.mo.momz
			
			if p.maril_postswimadjust then
				if p.aiming ~= 0 then
					p.aiming = $ - $/17
				end
				
				if p.mo.angle ~= p.drawangle then
					p.drawangle = $ - ($ - p.drawangle)/17
				end
				
				p.maril_postswimadjust = $-1
			end
		
			if p.maril_fly then
				marilFly(p)
			else
				marilSpin(p)
			end
		end
	end
	
	p.specki_laststate = p.mo.state
	p.specki_lastmomz = p.mo.momz
end)